# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General mutable functions for evolution and symbolic regression.

While users can use hyper primitives such as ``pg.oneof`` to fine-tune specific
parts of pre-existing programs, there are other use cases where it's necessary
to create a program from scratch or transform an existing program into something
drastically different. For example, one may need to find an activation formula
for neural networks or symbolically regress a function with only a few
observations. To address these needs, pg.mutfun has been introduced.

``pg.mutfun`` provides functions and instructions represented by symbolic
objects that allow for maximum flexibility in manipulating a program.
This includes but is not limited to inserting new lines, deleting existing ones,
replacing operations, and creating new functions. With ``pg.mutfun``, users have
access to APIs that make these tasks more manageable. They can easily identify
all downstream instructions that depend on the current instruction, or all
upstream instructions that the current instruction depends on, or finding out
all defined variables up to current instruction.

A mutfun program is a :class:`Function <pyglove.mutfun.Function>` object
illustrated as below::

  f = pg.mutfun.Function('f',
      [
          pg.mutfun.Assign('y', pg.mutfun.Var('x') + 1)
          pg.mutfun.Assign('z', pg.mutfun.Var('x') ** 2)
          pg.mutfun.Var('y') * pg.mutfun.Var('z')
      ], args=['x'])

  assert f(2) == (2 + 1) * 2 ** 2
  print(f)

  >> def f(x):
  >>   y = x + 1
  >>   z = x ** 2
  >>   return y + z

:doc:`../../../notebooks/evolution/function_regression` provides an example for
evolving and doing symbolic regression on mutable functions.

Class hierarchy:

.. graphviz::
   :align: center

    digraph codetypes {
      node [shape="box"];
      edge [arrowtail="empty" arrowhead="none" dir="back" style="dashed"];

      code [label="Code" href="code.html"]
      symbol_def [label="SymbolDefinition" href="symbol_definition.html"];
      assign [label="Assign" href="assign.html"];
      function [label="Function" href="function.html"];
      instruction [label="Instruction" href="instruction.html"];
      symbol_ref [label="SymbolReference" href="symbol_reference.html"];
      var [label="Var" href="var.html"];
      function_call [label="FunctionCall" href="function_call.html"]
      operator [label="Operator", href="operator.html"]
      unary_operator [label="UnaryOperator", href="unary_operator.html"]
      binary_operator [label="BinaryOperator", href="binary_operator.html"]
      user_defined [label="<User-defined instructions>"]

      code -> symbol_def;
      code -> instruction;
      symbol_def -> assign;
      symbol_def -> function;
      instruction -> symbol_ref;
      instruction -> operator;
      instruction -> user_defined;
      symbol_ref -> var;
      symbol_ref -> function_call;
      operator -> unary_operator;
      operator -> binary_operator;
      function -> code [arrowtail="diamond" style="none" label="body"];
    }
"""

# pylint: disable=g-bad-import-order

# Base and common instructions.
from pyglove.ext.mutfun.base import Code

from pyglove.ext.mutfun.base import SymbolDefinition
from pyglove.ext.mutfun.base import Function
from pyglove.ext.mutfun.base import Assign

from pyglove.ext.mutfun.base import Instruction
from pyglove.ext.mutfun.base import SymbolReference
from pyglove.ext.mutfun.base import Var
from pyglove.ext.mutfun.base import FunctionCall

from pyglove.ext.mutfun.base import evaluate
from pyglove.ext.mutfun.base import python_repr

# Basic operators.
from pyglove.ext.mutfun.basic_ops import Operator
from pyglove.ext.mutfun.basic_ops import UnaryOperator
from pyglove.ext.mutfun.basic_ops import Negate
from pyglove.ext.mutfun.basic_ops import BinaryOperator
from pyglove.ext.mutfun.basic_ops import Add
from pyglove.ext.mutfun.basic_ops import Substract
from pyglove.ext.mutfun.basic_ops import Multiply
from pyglove.ext.mutfun.basic_ops import Divide
from pyglove.ext.mutfun.basic_ops import FloorDivide
from pyglove.ext.mutfun.basic_ops import Mod
from pyglove.ext.mutfun.basic_ops import Power
from pyglove.ext.mutfun.basic_ops import Equals
from pyglove.ext.mutfun.basic_ops import NotEquals
from pyglove.ext.mutfun.basic_ops import GreaterThan
from pyglove.ext.mutfun.basic_ops import LessThan


# pylint: enable=g-bad-import-order



# Copyright 2019 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyGlove's genetic computing framework."""

from pyglove.ext import scalars
from pyglove.ext.evolution import base
from pyglove.ext.evolution import hill_climb as hill_climb_lib
from pyglove.ext.evolution import mutators
from pyglove.ext.evolution import neat as neat_lib
from pyglove.ext.evolution import nsga2 as nsga2_lib
from pyglove.ext.evolution import recombinators
from pyglove.ext.evolution import regularized_evolution as regularized_evolution_lib
from pyglove.ext.evolution import selectors
from pyglove.ext.evolution.hill_climb import hill_climb
from pyglove.ext.evolution.neat import neat
from pyglove.ext.evolution.nsga2 import nsga2
from pyglove.ext.evolution.regularized_evolution import regularized_evolution

# Alias for backward compatibility.
# Remove once dependencies are fixed.
RegularizedEvolution = regularized_evolution

# Interfaces.
Operation = base.Operation
DNAOperation = base.DNAOperation
Selector = base.Selector
Recombinator = base.Recombinator
Mutator = base.Mutator
Scalar = scalars.Scalar

# The compositional evolution class.
Evolution = base.Evolution

# Compositional operators.
# Common operations that are not associated with an operator.
Lambda = base.Lambda

Pipeline = base.Pipeline                         # operator >>
Power = base.Power                               # operator **

Concatenation = base.Concatenation                # operator +
Slice = base.Slice                                # operator []
Repeat = base.Repeat                              # operator *

Identity = base.Identity
Union = base.Union                                # operator |
Intersection = base.Intersection                  # operator &
Difference = base.Difference                      # operator -
SymmetricDifference = base.SymmetricDifference    # operator ^
Inversion = base.Inversion                        # operator ~

Choice = base.Choice                             # .with_prob
Conditional = base.Conditional                   # .if_true/.if_false
ElementWise = base.ElementWise                   # .for_each
Flatten = base.Flatten                           # .flatten
UntilChange = base.UntilChange                   # .until_change


GlobalStateGetter = base.GlobalStateGetter       # .global_state
GlobalStateSetter = base.GlobalStateSetter       # .as_global_state

# Helper method.
scalar_spec = scalars.scalar_spec
scalar_value = scalars.scalar_value

set_fitness = base.set_fitness
get_fitness = base.get_fitness

get_generation_id = base.get_generation_id
get_feedback_sequence_number = base.get_feedback_sequence_number
is_initial_population = base.is_initial_population





# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest

from pyglove.core import symbolic   # pylint: disable=unused-import
from pyglove.core.views.html.controls import progress_bar


class ProgressBarTest(unittest.TestCase):

  def assert_html_content(self, control, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = control.to_html().content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_basic(self):
    bar = progress_bar.ProgressBar(
        subprogresses=[
            progress_bar.SubProgress('foo'),
            progress_bar.SubProgress('bar', 20),
        ],
        total=None,
    )
    self.assert_html_content(
        bar,
        (
            '<div class="progress-bar"><div class="shade">'
            f'<div class="sub-progress foo" id="{bar["foo"].element_id()}">'
            '</div><div class="sub-progress bar" '
            f'id="{bar["bar"].element_id()}"></div></div>'
            '<div class="label-container"><span class="label progress-label"'
            f' id="{bar._progress_label.element_id()}">n/a</span><span class='
            f'"tooltip" id="{bar._progress_label.tooltip.element_id()}">'
            'Not started</span></div></div>'
        )
    )
    self.assertEqual(bar['foo'], progress_bar.SubProgress('foo'))
    self.assertEqual(bar['bar'], progress_bar.SubProgress('bar', 20))
    with self.assertRaisesRegex(KeyError, 'Sub progress bar .* not found'):
      _ = bar['baz']
    self.assertIsNone(bar['foo'].total)
    self.assertIsNone(bar['foo'].width)

    bar.update(total=100)
    self.assertEqual(bar.total, 100)
    self.assertEqual(bar['foo'].total, 100)
    self.assertEqual(bar['foo'].width, '0%')
    self.assertEqual(bar['bar'].width, '20%')
    with bar.track_scripts() as scripts:
      bar['foo'].increment()
      self.assertEqual(
          scripts,
          [
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{bar['foo'].element_id()}");
                  elem.style = "width:1%;";
                  """
              ),
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{bar._progress_label.element_id()}");
                  elem.textContent = " 21.0% (21/100)";
                  """
              ),
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{bar._progress_label.tooltip.element_id()}");
                  elem.textContent = "foo: 1.0% (1/100)\\nbar: 20.0% (20/100)";
                  """
              ),
              inspect.cleandoc(
                  f"""
                  elem = document.getElementById("{bar._progress_label.tooltip.element_id()}");
                  elem.classList.remove("html-content");
                  """
              ),
          ]
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest

from pyglove.core import symbolic   # pylint: disable=unused-import
from pyglove.core.views.html import base
from pyglove.core.views.html.controls import tab as tab_lib


class TabControlTest(unittest.TestCase):

  def assert_html_content(self, control, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = control.to_html().content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_find_tab_index(self):
    self.assertEqual(
        tab_lib.TabControl._find_tab_index(
            [
                tab_lib.Tab('foo', base.Html('<h1>foo</h1>')),
                tab_lib.Tab('bar', base.Html('<h1>bar</h1>')),
            ],
            1
        ),
        1
    )
    self.assertEqual(
        tab_lib.TabControl._find_tab_index(
            [
                tab_lib.Tab('foo', base.Html('<h1>foo</h1>'), name='foo'),
                tab_lib.Tab('bar', base.Html('<h1>bar</h1>'), name='bar'),
            ],
            'bar'
        ),
        1
    )
    self.assertEqual(
        tab_lib.TabControl._find_tab_index(
            [
                tab_lib.Tab('foo', base.Html('<h1>foo</h1>'), name='foo'),
                tab_lib.Tab('bar', base.Html('<h1>bar</h1>'), name='bar'),
            ],
            'baz'
        ),
        -1
    )
    self.assertEqual(
        tab_lib.TabControl._find_tab_index(
            [
                tab_lib.Tab('foo', base.Html('<h1>foo</h1>'), name='foo'),
                tab_lib.Tab('bar', base.Html('<h1>bar</h1>'), name='bar'),
                tab_lib.Tab('baz', base.Html('<h1>baz</h1>'), name='baz'),
            ],
            ['baz', 'bar']
        ),
        2
    )

  def test_init(self):
    tab = tab_lib.TabControl(
        [
            [tab_lib.Tab('foo', base.Html('<h1>foo</h1>'), name='foo')],
            None,
            tab_lib.Tab('bar', base.Html('<h1>bar</h1>'), name='bar'),
        ],
        1
    )
    self.assertEqual(tab.selected, 1)
    self.assertEqual(tab.tabs[0].name, 'foo')
    self.assertEqual(tab.tabs[1].name, 'bar')

    # Cannot find tab by name.
    tab = tab_lib.TabControl(
        [
            tab_lib.Tab('foo', base.Html('<h1>foo</h1>'), name='foo'),
            tab_lib.Tab('bar', base.Html('<h1>bar</h1>'), name='bar'),
        ],
        'baz'
    )
    self.assertEqual(tab.selected, 0)
    self.assertEqual(tab.tabs[0].name, 'foo')
    self.assertEqual(tab.tabs[1].name, 'bar')

  def test_basic(self):
    tab = tab_lib.TabControl([
        tab_lib.Tab(
            'foo', base.Html('<h1>foo</h1>'), css_classes=['foo'], name='foo'
        ),
        None,
        [tab_lib.Tab('bar', base.Html('<h1>bar</h1>'), name='bar')],
    ], selected='bar')
    self.assertEqual(tab.selected, 1)
    elem_id = tab.element_id()
    self.assert_html_content(
        tab,
        f"""<table class="tab-control"><tr><td><div class="tab-button-group top" id="{elem_id}-button-group"><button class="tab-button foo" onclick="openTab(event, '{elem_id}', '{elem_id}-0')"><span class="label">foo</span></button><button class="tab-button selected" onclick="openTab(event, '{elem_id}', '{elem_id}-1')"><span class="label">bar</span></button></div></td></tr><tr><td><div class="tab-content-group top" id="{elem_id}-content-group"><div class="tab-content foo" id="{elem_id}-0"><h1>foo</h1></div><div class="tab-content selected" id="{elem_id}-1"><h1>bar</h1></div></div></td></tr></table>"""
    )
    with tab.track_scripts() as scripts:
      tab.extend([
          tab_lib.Tab(
              'baz',
              base.Html(
                  '<h1 class="a">baz</h1>').add_style('.a { color: red; }')
              ),
          tab_lib.Tab('qux', base.Html('<h1>qux</h1>')),
      ])
      self.assertEqual(len(scripts), 6)
    with tab.track_scripts() as scripts:
      tab.insert(0, tab_lib.Tab('x', 'foo', name='x'))
      self.assertEqual(len(scripts), 2)
      self.assertEqual(len(tab.tabs), 5)
    self.assertEqual(tab.indexof(-1), 4)
    self.assertEqual(tab.indexof(3), 3)
    self.assertEqual(tab.indexof(10), 4)
    self.assertEqual(tab.indexof(-10), -1)
    self.assertEqual(tab.indexof('x'), 0)
    self.assertEqual(tab.indexof('y'), -1)
    self.assertEqual(tab.select(0), 0)
    self.assertEqual(tab.select('x'), 'x')
    self.assertEqual(tab.select(['y', 'x']), 'x')

    with self.assertRaisesRegex(ValueError, 'Tab not found'):
      tab.select('y')
    with self.assertRaisesRegex(ValueError, 'Tab not found'):
      tab.insert('y', tab_lib.Tab('z', 'bar'))

    with tab.track_scripts() as scripts:
      tab.insert('x', tab_lib.Tab('y', 'bar'))
      self.assertEqual(len(scripts), 2)
      self.assertEqual(len(tab.tabs), 6)

    with tab.track_scripts() as scripts:
      tab.select(3)
      self.assertEqual(len(scripts), 1)
      self.assertEqual(tab.selected, 3)

    with tab.track_scripts() as scripts:
      tab.select('x')
      self.assertEqual(len(scripts), 1)
      self.assertEqual(tab.selected, 1)


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import random
import unittest

from pyglove.core import geno
from pyglove.core import symbolic
from pyglove.core import utils
from pyglove.core.hyper.categorical import oneof
from pyglove.core.hyper.custom import CustomHyper
from pyglove.core.hyper.iter import iterate
from pyglove.core.hyper.object_template import materialize


class IntSequence(CustomHyper):

  def custom_decode(self, dna):
    return [int(v) for v in dna.value.split(',')]


class IntSequenceWithEncode(IntSequence):

  def custom_encode(self, value):
    return geno.DNA(','.join([str(v) for v in value]))

  def next_dna(self, dna):
    if dna is None:
      return geno.DNA(','.join([str(i) for i in range(5)]))
    v = self.custom_decode(dna)
    v.append(len(v))
    return self._create_dna(v)

  def random_dna(self, random_generator, previous_dna):
    del previous_dna
    k = random_generator.randint(0, 10)
    v = random_generator.choices(list(range(10)), k=k)
    return self._create_dna(v)

  def _create_dna(self, numbers):
    return geno.DNA(','.join([str(n) for n in numbers]))


class CustomHyperTest(unittest.TestCase):
  """Test for CustomHyper."""

  def test_dna_spec(self):
    self.assertTrue(
        symbolic.eq(
            IntSequence(hints='x').dna_spec('a'),
            geno.CustomDecisionPoint(
                hyper_type='IntSequence', location=utils.KeyPath('a'), hints='x'
            ),
        )
    )

  def test_decode(self):
    self.assertEqual(IntSequence().decode(geno.DNA('0,1,2')), [0, 1, 2])
    self.assertEqual(IntSequence().decode(geno.DNA('0')), [0])
    with self.assertRaisesRegex(ValueError, '.* expects string type DNA'):
      IntSequence().decode(geno.DNA(1))

  def test_encode(self):
    self.assertEqual(
        IntSequenceWithEncode().encode([0, 1, 2]), geno.DNA('0,1,2'))

    with self.assertRaisesRegex(
        NotImplementedError, '\'custom_encode\' is not supported by'):
      _ = IntSequence().encode([0, 1, 2])

  def test_random_dna(self):
    self.assertEqual(
        geno.random_dna(
            IntSequenceWithEncode().dna_spec('a'), random.Random(1)),
        geno.DNA('5,8'))

    with self.assertRaisesRegex(
        NotImplementedError, '`random_dna` is not implemented in .*'):
      geno.random_dna(IntSequence().dna_spec('a'))

  def test_iter(self):
    self.assertEqual(IntSequenceWithEncode().first_dna(), geno.DNA('0,1,2,3,4'))
    self.assertEqual(
        list(iterate(IntSequenceWithEncode(), 3)),
        [[0, 1, 2, 3, 4],
         [0, 1, 2, 3, 4, 5],
         [0, 1, 2, 3, 4, 5, 6]])

    with self.assertRaisesRegex(
        NotImplementedError, '`next_dna` is not implemented in .*'):
      next(iterate(IntSequence()))

  def test_interop_with_other_primitives(self):
    v = oneof([IntSequence(), 1, 2])
    self.assertEqual(materialize(v, geno.DNA(1)), 1)
    self.assertEqual(materialize(v, geno.DNA((0, '3,4'))), [3, 4])


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest

from pyglove.core.coding import errors
from pyglove.core.coding import execution


def code_error(code: str) -> errors.CodeError:
  try:
    execution.run(inspect.cleandoc(code), timeout=2)
    assert False, 'should not reach here'
  except errors.CodeError as e:
    return e


class CodeErrorsTest(unittest.TestCase):

  def test_format(self):
    e = code_error(
        """
        x = y + 1
        """
    )
    self.assertIn('[Code]', str(e))
    self.assertNotIn(
        '[Code]', e.format(include_complete_code=False))

  def test_lineno(self):
    self.assertEqual(
        code_error(
            """
            x = y + 1
            """
        ).lineno, 1)
    self.assertEqual(
        code_error(
            """
            x = 1
            for i of x:
              y = i
            """
        ).lineno, 2)
    self.assertEqual(
        code_error(
            """
            x = 1
            y = 2
            raise ValueError
            """
        ).lineno, 3)

  def test_lineno_in_error_message(self):
    def assert_lineno(code):
      e = code_error(code)
      self.assertIn('line', e.format(include_complete_code=False))

    assert_lineno(
        """
        x = y + 1
        """
    )
    assert_lineno(
        """
        x = 1
          y = 2
        """
    )
    assert_lineno(
        """
        raise ValueError()
        """
    )


class SerializationErrorTest(unittest.TestCase):

  def test_str(self):
    e = errors.SerializationError(
        'Output cannot be serialized.', ValueError('abc'))
    self.assertIn('Output cannot be serialized', str(e))
    self.assertIn('ValueError: abc', str(e))


if __name__ == '__main__':
  unittest.main()
